package u7a2;

import java.util.Vector;

/**
 * Utility functions for binary search trees containing generic things
 *
 * @param <T> The type of the things which are managed in the binary search tree
 */
public interface IBinarySearchTreeUtils<T> {
	/**
	 * Determines the height of a tree.
	 * 
	 * The empty tree has height zero. 
	 * The tree consisting only of the root has height one.
	 * 
	 * @param tree the tree in question
	 * @return the height of the given tree
	 */
	public int height(BinarySearchTree<T> tree);

	/**
	 * Checks if a tree is actually a leaf node.
	 * 
	 * A tree is a leave node iff it has no children.
	 * 
	 * @param tree thre tree in question. May not be null.
	 * @return true iff the given tree is a leaf node.
	 */
	public boolean isLeaf(BinarySearchTree<T> tree);
	
	/**
	 * Checks if a tree has only one child.
	 * 
	 * @param tree the tree in question
	 * @return true iff the given tree has only one child.
	 */
	public boolean hasOneChild(BinarySearchTree<T> tree);

	/**
	 * Returns the objects of a tree in pre-order.
	 * 
	 * @param tree the tree containing the objects
	 * @return a vector of the objects of the given tree, sorted in pre-order.
	 */
	public Vector<T> preOrder(BinarySearchTree<T> tree);
	
	/**
	 * Returns the objects of a tree in in-order.
	 *  
	 * @param tree the tree containing the objects
	 * @return a vector of the objects of the given tree, sorted in in-order.
	 */
	public Vector<T> inOrder(BinarySearchTree<T> tree);
	
	/**
	 * Returns the objects of a tree in post-order.
	 * 
	 * @param tree the tree containing the objects
	 * @return a vector of the objects of the given tree, sorted in post-order.
	 */
	public Vector<T> postOrder(BinarySearchTree<T> tree);
	
    /**
     * Inserts an object into a tree
     * 
     * If an object with the given key already exists in the tree it is overwritten.
     * 
     * @param tree the tree into which the object is inserted.
     *     After the call this tree should not be used any more.
     *     Use the returned tree instead.
     * @param key the key of the object
     * @param object the object which is inserted into the tree
     * @return a new tree with the given object inserted.
     */
	public BinarySearchTree<T> insert(BinarySearchTree<T> tree, int key, T object);
	
	/**
	 * Finds an object in a tree.
	 * 
	 * @param tree the tree in which the object is searched for.
	 * @param key the key of the object which is searched
	 * @return the object if it exists in the tree, and null otherwise.
	 */
	public Object find(BinarySearchTree<T> tree, int key);
	
	/**
	 * Result or the unlinkSmallest operation
	 */
	class UnlinkSmallestResult<T> {
		/**
		 * the tree with the smallest key
		 */
		public BinarySearchTree<T> smallest;
		
		/**
		 * the remaining tree with the smallest tree removed
		 */
		public BinarySearchTree<T> tree;
		
		public UnlinkSmallestResult(BinarySearchTree<T> smallest, BinarySearchTree<T> tree) {
			this.smallest = smallest;
			this.tree = tree;
		}
	}
	
	/**
	 * Removes and returns the sub-tree with the smallest key.
	 * 
	 * @param tree the tree in question
	 *     After this call this tree should not be used any more.
	 *     Use the returned {@link UnlinkSmallestResult#tree} instead.
	 * @return both the tree with the smallest key and the remaining tree with the smallest removed
	 */
	public UnlinkSmallestResult<T> unlinkSmallest(BinarySearchTree<T> tree);
	
	/**
	 * Removes an object from a tree.
	 * 
	 * The removed node is replaced with the smallest node of its right subtree.
	 * 
	 * @param tree the tree from which the object is removed.
	 *     After this call this tree should not be used any more.
	 *     Use the returned tree instead.
	 * @param key the key of the object
	 * @return either a new tree with the given object removed 
	 *     or the simpy the given tree if the object could not be found.
	 */
	public BinarySearchTree<T> remove(BinarySearchTree<T> tree, int key);
}